﻿#region Copyright

// A²DS - Autonomous Aerial Drone Simulator
// http://anjdreas.spaces.live.com/
//  
// A Master of Science thesis on autonomous flight at the 
// Norwegian University of Science and Technology (NTNU).
//  
// Copyright © 2009-2010 by Andreas Larsen.  All rights reserved.

#endregion

#region Using

using System;
using System.Collections.Generic;

#endregion

namespace Anj.Helpers.XNA
{
    /// <summary>
    ///   A ring variant of the generic list.
    ///   The main difference is a max capacity that will never be exceeded and
    ///   that the items are ordered from the most recent to the least recent (inverse of normal list).
    /// </summary>
    /// <typeparam name="T"></typeparam>
    public class RingBuffer<T> : List<T>
    {
        private readonly int _ringCapacity;
        private readonly object _syncRoot;

        public RingBuffer(int ringCapacity)
            : base(ringCapacity)
        {
            _syncRoot = new object();
            _ringCapacity = ringCapacity;
        }

        /// <summary>
        ///   Override for ring buffer behavior. Adds items to the 0-index.
        /// </summary>
        /// <param name="item"></param>
        public new void Add(T item)
        {
            lock (_syncRoot)
            {
                if (Count > _ringCapacity)
                    throw new Exception("Ring buffer has somehow exceeded its limit?");

                // Limit ring to its capacity by removing oldest item
                if (Count == _ringCapacity)
                    RemoveAt(Count - 1);

                // Then add the latest item to the queue
                Insert(0, item);
            }
        }

        /// <summary>Not supported.</summary>
        /// <param name="items"></param>
        public new void AddRange(IEnumerable<T> items)
        {
            throw new NotSupportedException(
                "Use Add() instead, because it is not intuitive what order the items will be added.");
        }
    }
}