﻿#region Copyright

// A²DS - Autonomous Aerial Drone Simulator
// http://anjdreas.spaces.live.com/
//  
// A Master of Science thesis on autonomous flight at the 
// Norwegian University of Science and Technology (NTNU).
//  
// Copyright © 2009-2010 by Andreas Larsen.  All rights reserved.

#endregion

#region Using

using System;
using Microsoft.Xna.Framework;

#endregion

namespace Anj.Helpers.XNA
{
    public static class VectorHelper
    {
        // Let heading of 0 degrees point towards north (Vector3.Forward)
        private static readonly Vector2 BaseVector2 = Vector2.UnitY; //Vector2.UnitX;
        private static readonly GaussianRandom GaussRand;

        static VectorHelper()
        {
            GaussRand = new GaussianRandom();
        }

        public static string ToString(Vector3 v, int precision)
        {
            return "{X:" + Math.Round(v.X, precision) + " Y:" + Math.Round(v.Y, precision) + " Z:" +
                   Math.Round(v.Z, precision) + "}";
        }

        /// <summary>
        ///   Adds pitch, roll and yaw delta angles in radians to an existing rotation matrix 
        ///   around its corresponding Right, Up and Forward axes.
        ///   Example: Joystick inputs control the pitch, roll and yaw of the helicopter and
        ///   this method is used to rotate the helicopter accordingly.
        /// </summary>
        /// <param name="rotation"></param>
        /// <param name="pitch"></param>
        /// <param name="roll"></param>
        /// <param name="yaw"></param>
        /// <returns></returns>
        public static Quaternion AddPitchRollYaw(Quaternion rotation, float pitch, float roll, float yaw)
        {
            // TODO Take into account that the order of rotation matters
            // See issue #21. 

            // Calculate quaternions in order to rotate the helicopter around its local X,Y and Z axes
            Quaternion relativeRotX = Quaternion.CreateFromAxisAngle(Vector3.Right, pitch);
            Quaternion relativeRotY = Quaternion.CreateFromAxisAngle(Vector3.Up, yaw);
            Quaternion relativeRotZ = Quaternion.CreateFromAxisAngle(Vector3.Forward, roll);

            return rotation*relativeRotX*relativeRotY*relativeRotZ;
        }

        /// <summary>
        ///   Obtain a rotation matrix from three perpendicular unit vectors.
        /// </summary>
        /// <param name="axes">
        ///   Axes defined by three perpendicular unit vectors
        /// </param>
        /// <returns></returns>
        public static Matrix GetRotation(Axes axes)
        {
            return new Matrix
                       {
                           Forward = axes.Forward,
                           Right = axes.Right,
                           Up = axes.Up
                       };
        }

        /// <summary>
        ///   Returns vectors X,Y,Z in an array given a world rotation matrix.
        /// </summary>
        /// <param name="rotation"></param>
        /// <returns></returns>
        public static Axes GetAxes(Matrix rotation)
        {
            return new Axes(rotation.Forward, rotation.Right, rotation.Up);
        }


        public static Axes GetAxes(Quaternion orientation)
        {
            // We define forward as -Z, right as +X and up as +Y at identity rotation (pitch = roll = yaw = 0)
            Vector3 forward = Vector3.Transform(Vector3.Forward, orientation);
            Vector3 right = Vector3.Transform(Vector3.Right, orientation);
            Vector3 up = Vector3.Transform(Vector3.Up, orientation);
            return new Axes(forward, right, up);
        }

        public static Vector3 ToHorizontal3D(Vector3 v)
        {
            return HorizontalTo3D(ToHorizontal(v));
        }

        public static Vector2 ToHorizontal(Vector3 v)
        {
            // Invert Y-axis, because it is natural to think of negative Z translated to positive Y when seen from
            // a top view perspective
            return new Vector2(v.X, -v.Z);
        }

        public static Vector2 ToHorizontal(float angle)
        {
            // Invert Y-axis, because it is natural to think of negative Z translated to positive Y when seen from
            // a top view perspective
            return new Vector2((float) Math.Cos(angle), -(float) Math.Sin(angle));
        }

        public static Vector3 HorizontalTo3D(Vector2 horizontalVector)
        {
            // Invert Y-axis, because it is natural to think of negative Z translated to positive Y when seen from
            // a top view perspective
            return new Vector3(horizontalVector.X, 0, -horizontalVector.Y);
        }

        public static float GetAngle(Vector2 v)
        {
            return GetAngle(v, BaseVector2);
        }

        public static float GetAngle(Vector2 v, Vector2 reference)
        {
//            var dot = MathHelper.Clamp(Vector2.Dot(Vector2.Normalize(v), Vector2.Normalize(reference)), -1, +1);
//            var result = (float) Math.Acos(dot);
            var angle = (float) Math.Atan2(v.Y, v.X);
            var refAngle = (float) Math.Atan2(reference.Y, reference.X);
            return DiffAngle(angle, refAngle);
        }

        public static float GetAngle(Vector3 a, Vector3 b)
        {
            float dot = MathHelper.Clamp(Vector3.Dot(Vector3.Normalize(a), Vector3.Normalize(b)), -1, +1);
            return (float) Math.Acos(dot);
        }

        public static float GetPitchAngle(Quaternion orientation)
        {
            return GetPitchAngle(Matrix.CreateFromQuaternion(orientation));
        }

        /// <summary>
        /// http://planning.cs.uiuc.edu/node103.html
        /// http://en.wikipedia.org/wiki/Euler_angles#Table_of_matrices using XNA's YXZ convention
        /// </summary>
        /// <param name="cdm"></param>
        /// <returns></returns>
        public static float GetPitchAngle(Matrix cdm)
        {
            //if (Math.Abs(cdm.M32) != 1)
            //{
            double pitch = -Math.Asin(cdm.M32);
//            double yaw = -Math.Atan2(cdm.M31, cdm.M33);
//            double roll = -Math.Atan2(cdm.M12, cdm.M22);
            return (float) pitch;
            //}
            //else
            //{

            //}

            // Positive pitch as nose towards the sky
            return (float) Math.Atan2(cdm.Forward.Z, cdm.Forward.Y);
        }

        public static float GetRollAngle(Quaternion orientation)
        {
            return GetRollAngle(Matrix.CreateFromQuaternion(orientation));
        }

        /// <summary>
        /// http://planning.cs.uiuc.edu/node103.html
        /// http://en.wikipedia.org/wiki/Euler_angles#Table_of_matrices using XNA's YXZ convention 
        /// </summary>
        /// <param name="cdm"></param>
        /// <returns></returns>
        public static float GetRollAngle(Matrix cdm)
        {
            //            double yaw = -Math.Atan2(cdm.M31, cdm.M33);
            //            double pitch = -Math.Asin(cdm.M32);
            double roll = -Math.Atan2(cdm.M12, cdm.M22);
            return (float) roll;

            // Positive roll when rolling to the "right" as seen from the cockpit
            return (float) Math.Atan2(cdm.Up.X, cdm.Up.Y);
            //return -(float)Math.Atan2(-cdm.M21, Math.Sqrt(cdm.M22 * cdm.M22 + cdm.M23 * cdm.M23));
        }

        public static float GetHeadingAngle(Quaternion orientation)
        {
            return -GetHeadingAngle(Matrix.CreateFromQuaternion(orientation));
        }

        /// <summary>
        /// http://planning.cs.uiuc.edu/node103.html
        /// http://en.wikipedia.org/wiki/Euler_angles#Table_of_matrices using XNA's YXZ convention
        /// </summary>
        /// <param name="cdm"></param>
        /// <returns></returns>
        public static float GetHeadingAngle(Matrix cdm)
        {
//            double pitch = -Math.Asin(cdm.M32);
            double yaw = -Math.Atan2(cdm.M31, cdm.M33);
//            double roll = -Math.Atan2(cdm.M12, cdm.M22);
            return (float) yaw;

            // Negate to get positive heading angle as yawing to the "right" in level flight
            return -(float) Math.Atan2(cdm.Forward.X, cdm.Right.X);
        }

        public static float GetHeadingAngle(Vector3 forward)
        {
            return GetAngle(ToHorizontal(forward));
        }

        /// <summary>
        ///   Return the difference between an angle and a reference angle.
        ///   If the angle is clockwise less than 180 degrees, the result is positive.
        ///   If the angle is counter-clockwise less than 180 degrees, the result is negative.
        ///   If the angle is exactly 180 degrees, the result is positive 180 degrees.
        /// </summary>
        /// <param name="angle"></param>
        /// <param name="reference"></param>
        /// <returns></returns>
        public static float DiffAngle(float angle, float reference)
        {
            float diff = angle - reference;
            var result = (float) Math.Atan2(Math.Sin(diff), Math.Cos(diff));

            // Handle the special case where the angles are close to or exactly 180 degrees apart.
            // The function would return 0 degrees, but this is just a mathematical anomaly.
            if (result == 0 &&
                MathHelper.Distance(angle, reference) > MathHelper.PiOver2)
                return MathHelper.Pi;

            return result;
        }

        /// <summary>
        ///   Returns the projection length of vector v onto a reference vector.
        /// </summary>
        /// <param name="v"></param>
        /// <param name="reference"></param>
        /// <returns></returns>
        public static float Project(Vector2 v, Vector2 reference)
        {
            if (reference.Length() < 0.0001f) return 0;

            return Vector2.Dot(v, reference)/reference.Length();
        }

        /// <summary>
        ///   Returns the projection length of vector v onto a reference vector.
        /// </summary>
        /// <param name="v"></param>
        /// <param name="reference"></param>
        /// <returns></returns>
        public static float Project(Vector3 v, Vector3 reference)
        {
            if (reference.Length() < 0.0001f) return 0;

            return Vector3.Dot(v, reference)/reference.Length();
        }

        /// <summary>
        ///   Returns a 2D vector perpendicular to the input vector.
        /// </summary>
        /// <param name="v"></param>
        /// <returns></returns>
        public static Vector2 CrossProduct(Vector2 v)
        {
            return new Vector2(v.Y, -v.X);
        }

        /// <summary>
        ///   Returns a Vector3 of random direction and radius between min and max.
        /// </summary>
        /// <param name="minRadius"></param>
        /// <param name="maxRadius"></param>
        /// <returns></returns>
        public static Vector3 Random3(float minRadius, float maxRadius)
        {
            if (minRadius < 0) throw new ArgumentException("Must be positive", "minRadius");
            if (maxRadius < 0) throw new ArgumentException("Must be positive", "maxRadius");
            if (minRadius == maxRadius) throw new ArgumentException("minRadius and maxRadius can't be equal.");

            var rand = new Random();

            // Use a random radius between 
            float randomRadius = MyMathHelper.Lerp((float) rand.NextDouble(), 0, 1, minRadius, maxRadius);

            // Random range from -1 to +1 for each axis
            var randomDirection = new Vector3(
                (float) (2*rand.NextDouble() - 1),
                (float) (2*rand.NextDouble() - 1),
                (float) (2*rand.NextDouble() - 1));

            // Cut length of vector to radius
            return Vector3.Normalize(randomDirection)*randomRadius;
        }

        /// <summary>
        ///   Returns a Vector3 of random direction and radius between min and max.
        /// </summary>
        /// <returns></returns>
        public static Vector3 GaussRandom3(Vector3 stdDev)
        {
            if (stdDev.X < 0 || stdDev.Y < 0 || stdDev.Z < 0)
                throw new ArgumentException("XYZ must be positive", "stdDev");

            return new Vector3(
                GaussRand.NextGaussian(0, stdDev.X),
                GaussRand.NextGaussian(0, stdDev.Y),
                GaussRand.NextGaussian(0, stdDev.Z));
        }

//        /// <summary>
//        /// Map a vector in a rotated coordinate system (not sure if translation/scaling will work as expected) to world cartesian coordinates.
//        /// Example: Map body frame (forward, right, up) acceleration vector to world acceleration vector (XYZ).
//        /// Note there is a precision issue when transforming a vector between coordinate systems.
//        /// </summary>
//        /// <param name="v"></param>
//        /// <param name="fromReferenceAxes"></param>
//        /// <returns></returns>
//        public static Vector3 MapToWorld(Vector3 v, Axes fromReferenceAxes)
//        {
//            return Vector3.Transform(v, )
////            return new Vector3
////                       {
////                           X = v.X*fromReferenceAxes.Right.X +
////                               v.Y*fromReferenceAxes.Up.X +
////                               v.Z*fromReferenceAxes.Forward.X,
////                           Y = v.X*fromReferenceAxes.Right.Y +
////                               v.Y*fromReferenceAxes.Up.Y +
////                               v.Z*fromReferenceAxes.Forward.Y,
////                           Z = v.X*fromReferenceAxes.Right.Z +
////                               v.Y*fromReferenceAxes.Up.Z +
////                               v.Z*fromReferenceAxes.Forward.Z
////                       };
//        }

        /// <summary>
        /// Map a vector in a rotated coordinate system (not sure if translation/scaling will work as expected) to world cartesian coordinates.
        /// Example: Map body frame (forward, right, up) acceleration vector to world acceleration vector (XYZ).
        /// Note there is a precision issue when transforming a vector between coordinate systems.
        /// </summary>
        /// <param name="v"></param>
        /// <param name="fromCoordinateSystem">The matrix should only consist of rotational operations, and no translation or scaling.</param>
        /// <returns></returns>
        public static Vector3 MapToWorld(Vector3 v, Matrix fromCoordinateSystem)
        {
//            return MapToWorld(v, new Axes(fromCoordinateSystem.Forward, fromCoordinateSystem.Right, fromCoordinateSystem.Up));
            return Vector3.Transform(v, Matrix.Invert(fromCoordinateSystem)); // Improved precision !
        }

        /// <summary>
        /// Map a vector in a rotated coordinate system (not sure if translation/scaling will work as expected) to world cartesian coordinates.
        /// Example: Map body frame (forward, right, up) acceleration vector to world acceleration vector (XYZ).
        /// Note there is a precision issue when transforming a vector between coordinate systems.
        /// </summary>
        /// <param name="v"></param>
        /// <param name="fromCoordinateSystem">The matrix should only consist of rotational operations, and no translation or scaling.</param>
        /// <returns></returns>
        public static Vector3 MapToWorld(Vector3 v, Quaternion fromCoordinateSystem)
        {
            //            return MapToWorld(v, new Axes(fromCoordinateSystem.Forward, fromCoordinateSystem.Right, fromCoordinateSystem.Up));
            return Vector3.Transform(v, Quaternion.Inverse(fromCoordinateSystem)); // Improved precision !
        }

        public static Vector3 MapFromWorld(Vector3 vectorWorld, Quaternion toCoordinateSystem)
        {
            return Vector3.Transform(vectorWorld, toCoordinateSystem);
        }
    }
}