﻿#region Copyright

// A²DS - Autonomous Aerial Drone Simulator
// http://anjdreas.spaces.live.com/
//  
// A Master of Science thesis on autonomous flight at the 
// Norwegian University of Science and Technology (NTNU).
//  
// Copyright © 2009-2010 by Andreas Larsen.  All rights reserved.

#endregion

#region Using

using System;
using System.Collections.Generic;
using System.IO;
using System.Xml.XPath;
using Anj.Helpers.XNA;
using Control;
using Control.Common;
using Microsoft.Xna.Framework;
using Simulator.Cameras;
using Simulator.Scenarios;
using Simulator.Utils;
using System.Diagnostics;
using Simulator.Testing;

#endregion

namespace Simulator.Parsers
{
    public struct TerrainInfo
    {
        public int Width;
        public float MinHeight;
        public float MaxHeight;
    }

    public static class ScenarioParserXML
    {
        public static string GetPreSelectedScenarioName(string xmlFilepath)
        {
            string xmlText = ParseHelper.GetResourceText(xmlFilepath);
            var reader = new StringReader(xmlText);

            var doc = new XPathDocument(reader);
            XPathNavigator nav = doc.CreateNavigator().SelectSingleNode("/root");
            XPathNavigator preSelectedScenarioNode = nav.SelectSingleNode("PreSelectedScenario");
            return preSelectedScenarioNode.Value;
        }


        public static IList<Scenario> GetScenarios(string xmlFilepath)
        {
            var result = new List<Scenario>();

            string xmlText = ParseHelper.GetResourceText(xmlFilepath);
            var reader = new StringReader(xmlText);

            var doc = new XPathDocument(reader);
            XPathNavigator nav = doc.CreateNavigator().SelectSingleNode("/root");
            XPathNodeIterator scenariosIter = nav.Select("Scenario");
            while (scenariosIter.MoveNext())
                result.Add(ParseScenario(scenariosIter.Current));

            reader.Close();

            return result;
        }

        private static Scenario ParseScenario(XPathNavigator scenarioNode)
        {
            var scenario = new Scenario();
            scenario.Name = scenarioNode.GetAttribute("Name");
            scenario.BackgroundMusic = ParseBackgroundMusic(scenarioNode.SelectSingleNode("BackgroundMusic"));
            scenario.SceneElements = ParseSceneElements(scenarioNode.SelectSingleNode("Scene"));
            scenario.CameraType = ParseCameraType(scenarioNode.SelectSingleNode("CameraType"));
            scenario.Timeout = ParseTimeout(scenarioNode.SelectSingleNode("TimeoutSeconds"));

            XPathNavigator terrainInfoNode = scenarioNode.SelectSingleNode("Scene").SelectSingleNode("Terrain");
            if (terrainInfoNode != null)
                scenario.TerrainInfo = ParseTerrainInfo(terrainInfoNode);
            

            scenario.HelicopterScenarios = ParseHelicopterScenario(scenarioNode);
            return scenario;
        }

        private static TimeSpan ParseTimeout(XPathNavigator timeoutSecondsNode)
        {
            if (timeoutSecondsNode == null)
                return TimeSpan.Zero;

            int timeoutSeconds = timeoutSecondsNode.ValueAsInt;
            return TimeSpan.FromSeconds(timeoutSeconds);
        }

        private static TerrainInfo ParseTerrainInfo(XPathNavigator terrainInfoNode)
        {
            var result = new TerrainInfo();

            if (!int.TryParse(terrainInfoNode.GetAttribute("Width"), out result.Width))
                result.Width = 256;

            if (!float.TryParse(terrainInfoNode.GetAttribute("MinHeight"), out result.MinHeight))
                result.MinHeight = 0;

            if (!float.TryParse(terrainInfoNode.GetAttribute("MaxHeight"), out result.MaxHeight))
                result.MaxHeight = 20;

            return result;
        }

        /// <summary>
        /// Returns a list of names of the scene elements.
        /// </summary>
        /// <param name="sceneNode"></param>
        /// <returns></returns>
        private static IEnumerable<string> ParseSceneElements(XPathNavigator sceneNode)
        {
            XPathNodeIterator sceneElements = sceneNode.SelectChildren(XPathNodeType.Element);

            var result = new List<string>();
            while (sceneElements.MoveNext())
            {
                string elementName = sceneElements.Current.Name;
                result.Add(elementName);
            }

            return result;
        }

        /// <summary>
        /// Returns the resource path for the background music sound effect file.
        /// Returns null if music tag was missing or not recognized.
        /// </summary>
        /// <param name="backgroundMusicNode"></param>
        /// <returns></returns>
        private static string ParseBackgroundMusic(XPathNavigator backgroundMusicNode)
        {
            if (backgroundMusicNode == null) 
            {
                Console.WriteLine(@"Could not find background music.");
                return null;
            }

            string music = backgroundMusicNode.Value;
            if (music == "Airwolf")
                return "Audio/Airwolf Theme";

            return null;
        }

        private static Task ParseTask(XPathNavigator helicopterNode)
        {
            try
            {
                var task = new Task();

                XPathNavigator taskNode = helicopterNode.SelectSingleNode("Task");

                var heightAboveGroundNode = taskNode.SelectSingleNode("HoldHeightAboveGround");
                task.HoldHeightAboveGround = (heightAboveGroundNode != null) ? heightAboveGroundNode.ValueAsInt : -1;

                XPathNavigator loopNode = taskNode.SelectSingleNode("Loop");
                task.Loop = (loopNode != null) ? loopNode.ValueAsBoolean : false;

                XPathNavigator defaultRadiusNode = taskNode.SelectSingleNode("DefaultWaypointRadius");
                float defaultWaypointRadius = (defaultRadiusNode != null)
                                           ? (float) defaultRadiusNode.ValueAsDouble
                                           : DefaultAutopilotConfiguration.WaypointRadius;

                XPathNodeIterator waypointIter = taskNode.Select("Waypoint");
                while (waypointIter.MoveNext())
                {
                    XPathNavigator radiusNode = waypointIter.Current.SelectSingleNode("Radius");

                    float radius = (radiusNode != null) 
                        ? (float) radiusNode.ValueAsDouble 
                        : defaultWaypointRadius;

                    Waypoint wp = ParseWaypoint(waypointIter.Current, radius);
                    if (wp != null)
                        task.AddWaypoint(wp);
                }

                return task;
            }
            catch (Exception e)
            {
                Console.WriteLine(@"Error parsing scenario XML. " + e);
                return null;
            }
        }

        private static Waypoint ParseWaypoint(XPathNavigator waypointNode, float waypointRadius)
        {
            // TODO Implement heading angle (also in XML file!)
            Vector3 position = XMLHelper.ParseVector3(waypointNode.SelectSingleNode("Position"));
            
            float heading = 0f;
            var headingNode = waypointNode.SelectSingleNode("HeadingAngle");
            if (headingNode != null)
                heading = (float) headingNode.ValueAsDouble;

            var type = (WaypointType) Enum.Parse(typeof (WaypointType), waypointNode.SelectSingleNode("Type").Value);
            var result = new Waypoint(position, heading, type, waypointRadius);

            var secondsToWaitNode = waypointNode.SelectSingleNode("SecondsToWait");
            if (secondsToWaitNode != null)
                result.SecondsToWait = (float)secondsToWaitNode.ValueAsDouble;

            return result;
        }

        private static IList<HelicopterScenario> ParseHelicopterScenario(XPathNavigator scenarioNode)
        {
            var result = new List<HelicopterScenario>();

            XPathNodeIterator helicoptersIter = scenarioNode.Select("Helicopter");
            while (helicoptersIter.MoveNext())
            {
                XPathNavigator heliNode = helicoptersIter.Current;

                var heliScenario = new HelicopterScenario();
                heliScenario.Task = ParseTask(heliNode);

                XPathNavigator playerControlledNode = heliNode.SelectSingleNode("PlayerControlled");
                heliScenario.PlayerControlled = (playerControlledNode != null) ? playerControlledNode.ValueAsBoolean : false;

                XPathNavigator playEngineSoundNode = heliNode.SelectSingleNode("EngineSound");
                heliScenario.EngineSound = (playEngineSoundNode != null) ? playEngineSoundNode.ValueAsBoolean : false;

                XPathNavigator startPosNode = heliNode.SelectSingleNode("StartPosition");
                heliScenario.StartPosition = XMLHelper.ParseVector3(startPosNode);

                var assistedAutopilotNode = heliNode.SelectSingleNode("AssistedAutopilot");
                heliScenario.AssistedAutopilot = (assistedAutopilotNode != null)
                                                     ? assistedAutopilotNode.ValueAsBoolean
                                                     : false;


                result.Add(heliScenario);
            }

            return result;
        }

        private static CameraType ParseCameraType(XPathNavigator cameraTypeNode)
        
        {
            if (cameraTypeNode == null)
            {
                Debug.WriteLine("Using chase camera as default.");
                return CameraType.Chase;
            }

            string cameraType = cameraTypeNode.Value;

            if (cameraType == "Cockpit")
                return CameraType.Cockpit;

            if (cameraType == "Chase")
                return CameraType.Chase;

            if (cameraType == "Fixed")
                return CameraType.Fixed;

            if (cameraType == "Free")
                return CameraType.Free;

            throw new ArgumentException("Unknown camera type " + cameraType);
            
        }

        
    }

    public struct SimulationSettings
    {
        public bool SwapStereo;
        public RenderModes RenderMode;
    }
}