﻿#region Copyright

// A²DS - Autonomous Aerial Drone Simulator
// http://anjdreas.spaces.live.com/
//  
// A Master of Science thesis on autonomous flight at the 
// Norwegian University of Science and Technology (NTNU).
//  
// Copyright © 2009-2010 by Andreas Larsen.  All rights reserved.

#endregion

#region Using

using System.Collections.Generic;
using System.IO;
using System.Xml.XPath;
using Microsoft.Xna.Framework;
using Sensors.Model;
using Simulator.Resources;
using Simulator.Testing;
using Simulator.Utils;
using State.Model;

#endregion

namespace Simulator.Parsers
{
    public static class TestConfigurationParserXML
    {
        public static TestConfiguration GetConfiguration(string xmlFilepath)
        {
            var result = new TestConfiguration();

            string xmlText = ParseHelper.GetResourceText(xmlFilepath);
            var reader = new StringReader(xmlText);

            var doc = new XPathDocument(reader);
            XPathNavigator nav = doc.CreateNavigator().SelectSingleNode("/root/TestConfiguration");

            // Parse configuration
            result.FlyBySensors = nav.SelectSingleNode("FlyBySensors").ValueAsBoolean;
            result.UsePerfectSensors = nav.SelectSingleNode("UsePerfectSensors").ValueAsBoolean;
            result.UseRangeFinder = nav.SelectSingleNode("UseRangeFinder").ValueAsBoolean;
            result.UseGPS = nav.SelectSingleNode("UseGPS").ValueAsBoolean;
            result.UseINS = nav.SelectSingleNode("UseINS").ValueAsBoolean;

            result.Sensors = ParseSensorSpecifications(nav.SelectSingleNode("SensorSpecifications"));
            result.MaxHVelocities = ParseMaxHVelocities(nav.Select("MaxHVelocity"));

            List<string> scenarioNames = ParseScenarioNames(nav.Select("ScenarioName"));

            foreach (string scenarioName in scenarioNames)
                result.TestScenarios.Add(SimulatorResources.GetScenario(scenarioName));

            return result;
        }

        private static SensorSpecifications ParseSensorSpecifications(XPathNavigator sensorsNode)
        {
            XPathNavigator accelerometerNode = sensorsNode.SelectSingleNode("Accelerometer");
            XPathNavigator accelStdDevNode = accelerometerNode.SelectSingleNode("NoiseStdDev");

            XPathNavigator gpsPosStdDevNode = sensorsNode.SelectSingleNode("GPSPositionAxisStdDev");
            XPathNavigator gpsVelStdDevNode = sensorsNode.SelectSingleNode("GPSVelocityAxisStdDev");

            var r = new SensorSpecifications
                        {
                            AccelerometerFrequency = (float)accelerometerNode.SelectSingleNode("Frequency").ValueAsDouble,

                            AccelerometerStdDev = new ForwardRightUp(
                                (float)accelStdDevNode.SelectSingleNode("Forward").ValueAsDouble,
                                (float)accelStdDevNode.SelectSingleNode("Right").ValueAsDouble,
                                (float)accelStdDevNode.SelectSingleNode("Up").ValueAsDouble),

                            GPSPositionStdDev = new Vector3((float) gpsPosStdDevNode.ValueAsDouble),
                            GPSVelocityStdDev = new Vector3((float) gpsVelStdDevNode.ValueAsDouble),
                            OrientationAngleNoiseStdDev = (float)sensorsNode.SelectSingleNode("OrientationAngleNoiseStdDev").ValueAsDouble,
                        };

            return r;
        }

        private static List<float> ParseMaxHVelocities(XPathNodeIterator velocityIter)
        {
            var result = new List<float>();
            while (velocityIter.MoveNext())
                result.Add((float) velocityIter.Current.ValueAsDouble);

            return result;
        }

        private static List<string> ParseScenarioNames(XPathNodeIterator scenariosIter)
        {
            var result = new List<string>();
            while (scenariosIter.MoveNext())
                result.Add(scenariosIter.Current.Value);

            return result;
        }
    }
}